@@ -293,6 +293,7 @@ static int cn_ranap_rx_paging_cmd(struct hnbgw_cnlink *cnlink,
 {
 	const char *errmsg;
 	struct hnb_context *hnb;
+	bool is_ps = cnlink->pool->domain == DOMAIN_PS;
 
 	errmsg = cnlink_paging_add_ranap(cnlink, imsg);
 	if (errmsg) {
@@ -305,34 +306,46 @@ static int cn_ranap_rx_paging_cmd(struct hnbgw_cnlink *cnlink,
 	/* FIXME: determine which HNBs to send this Paging command,
 	 * rather than broadcasting to all HNBs */
 	llist_for_each_entry(hnb, &g_hnbgw->hnb_list, list) {
+		if (!hnb->hnb_registered)
+			continue;
+		if (is_ps)
+			HNBP_CTR_INC(hnb->persistent, HNB_CTR_PS_PAGING_ATTEMPTED);
+		else
+			HNBP_CTR_INC(hnb->persistent, HNB_CTR_CS_PAGING_ATTEMPTED);
 		rua_tx_udt(hnb, data, len);
 	}
 
 	return 0;
 }
 
-static int cn_ranap_rx_initiating_msg(struct hnbgw_cnlink *cnlink,
-				      const struct osmo_scu_unitdata_param *unitdata,
-				      RANAP_InitiatingMessage_t *imsg,
-				      const uint8_t *data, unsigned int len)
+static int ranap_rx_udt_dl_initiating_msg(struct hnbgw_cnlink *cnlink,
+					  const struct osmo_scu_unitdata_param *unitdata,
+					  RANAP_InitiatingMessage_t *imsg,
+					  const uint8_t *data, unsigned int len)
 {
 	switch (imsg->procedureCode) {
 	case RANAP_ProcedureCode_id_Reset:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_RESET);
 		return cn_ranap_rx_reset_cmd(cnlink, unitdata, imsg);
 	case RANAP_ProcedureCode_id_Paging:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_PAGING);
 		return cn_ranap_rx_paging_cmd(cnlink, imsg, data, len);
 	case RANAP_ProcedureCode_id_OverloadControl: /* Overload ind */
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_OVERLOAD_IND);
 		break;
 	case RANAP_ProcedureCode_id_ErrorIndication: /* Error ind */
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_ERROR_IND);
 		break;
 	case RANAP_ProcedureCode_id_ResetResource: /* request */
 	case RANAP_ProcedureCode_id_InformationTransfer:
 	case RANAP_ProcedureCode_id_DirectInformationTransfer:
 	case RANAP_ProcedureCode_id_UplinkInformationExchange:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_UNSUPPORTED);
 		LOGP(DRANAP, LOGL_NOTICE, "Received unsupported RANAP "
 		     "Procedure %ld from CN, ignoring\n", imsg->procedureCode);
 		break;
 	default:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_UNKNOWN);
 		LOGP(DRANAP, LOGL_NOTICE, "Received suspicious RANAP "
 		     "Procedure %ld from CN, ignoring\n", imsg->procedureCode);
 		break;
@@ -340,20 +353,23 @@ static int cn_ranap_rx_initiating_msg(struct hnbgw_cnlink *cnlink,
 	return 0;
 }
 
-static int cn_ranap_rx_successful_msg(struct hnbgw_cnlink *cnlink,
-					RANAP_SuccessfulOutcome_t *omsg)
+static int ranap_rx_udt_dl_successful_msg(struct hnbgw_cnlink *cnlink,
+					  RANAP_SuccessfulOutcome_t *omsg)
 {
 	switch (omsg->procedureCode) {
 	case RANAP_ProcedureCode_id_Reset: /* Reset acknowledge */
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_RESET);
 		return cn_ranap_rx_reset_ack(cnlink, omsg);
 	case RANAP_ProcedureCode_id_ResetResource: /* response */
 	case RANAP_ProcedureCode_id_InformationTransfer:
 	case RANAP_ProcedureCode_id_DirectInformationTransfer:
 	case RANAP_ProcedureCode_id_UplinkInformationExchange:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_UNSUPPORTED);
 		LOGP(DRANAP, LOGL_NOTICE, "Received unsupported RANAP "
 		     "Procedure %ld from CN, ignoring\n", omsg->procedureCode);
 		break;
 	default:
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_RANAP_RX_UDT_UNKNOWN);
 		LOGP(DRANAP, LOGL_NOTICE, "Received suspicious RANAP "
 		     "Procedure %ld from CN, ignoring\n", omsg->procedureCode);
 		break;
@@ -362,19 +378,18 @@ static int cn_ranap_rx_successful_msg(struct hnbgw_cnlink *cnlink,
 }
 
 
-static int _cn_ranap_rx(struct hnbgw_cnlink *cnlink,
-			const struct osmo_scu_unitdata_param *unitdata,
-			RANAP_RANAP_PDU_t *pdu, const uint8_t *data, unsigned int len)
+static int hnbgw_ranap_rx_udt_dl(struct hnbgw_cnlink *cnlink,
+				 const struct osmo_scu_unitdata_param *unitdata,
+				 RANAP_RANAP_PDU_t *pdu, const uint8_t *data, unsigned int len)
 {
 	int rc;
 
 	switch (pdu->present) {
 	case RANAP_RANAP_PDU_PR_initiatingMessage:
-		rc = cn_ranap_rx_initiating_msg(cnlink, unitdata, &pdu->choice.initiatingMessage,
-						data, len);
+		rc = ranap_rx_udt_dl_initiating_msg(cnlink, unitdata, &pdu->choice.initiatingMessage, data, len);
 		break;
 	case RANAP_RANAP_PDU_PR_successfulOutcome:
-		rc = cn_ranap_rx_successful_msg(cnlink, &pdu->choice.successfulOutcome);
+		rc = ranap_rx_udt_dl_successful_msg(cnlink, &pdu->choice.successfulOutcome);
 		break;
 	case RANAP_RANAP_PDU_PR_unsuccessfulOutcome:
 		LOGP(DRANAP, LOGL_NOTICE, "Received unsupported RANAP "
@@ -407,7 +422,7 @@ static int handle_cn_ranap(struct hnbgw_cnlink *cnlink, const struct osmo_scu_un
 		return -1;
 	}
 
-	rc = _cn_ranap_rx(cnlink, unitdata, pdu, data, len);
+	rc = hnbgw_ranap_rx_udt_dl(cnlink, unitdata, pdu, data, len);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RANAP_PDU, pdu);
 
 	return rc;
@@ -849,11 +864,7 @@ int hnbgw_cnlink_start_or_restart(struct hnbgw_cnlink *cnlink)
 			return 0;
 		}
 		/* else cnlink->hnbgw_sccp_user stays NULL and is set up below. */
-		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "cs7 instance %u has no SCCP instance yet\n", ss7->cfg.id);
-
-		/* All SCCP instances should originate from this function. So if there is no hnbgw_sccp_user for the cs7
-		 * instance, then the cs7 instance should not have an SCCP instance yet. */
-		OSMO_ASSERT(!ss7->sccp);
+		LOG_CNLINK(cnlink, DCN, LOGL_DEBUG, "cs7 instance %u has no configured SCCP instance yet\n", ss7->cfg.id);
 	}
 
 	/* No SCCP instance yet for this ss7. Create it. If no address name is given that resolves to a
@@ -992,7 +1003,7 @@ struct hnbgw_cnlink *hnbgw_cnlink_select(struct hnbgw_context_map *map)
 		if (cnlink) {
 			LOG_MAP(map, DCN, LOGL_INFO, "CN link paging record selects %s %d\n", cnpool->peer_name,
 				cnlink->nr);
-			rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_SUBSCR_PAGED));
+			CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_SUBSCR_PAGED);
 			return cnlink;
 		}
 		/* If there is no match, go on with other ways */
@@ -1042,7 +1053,7 @@ struct hnbgw_cnlink *hnbgw_cnlink_select(struct hnbgw_context_map *map)
 			if (nri_matches_cnlink) {
 				LOG_NRI(LOGL_DEBUG, "NRI matches %s %d, but this %s is currently not connected\n",
 					cnpool->peer_name, cnlink->nr, cnpool->peer_name);
-				rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_SUBSCR_ATTACH_LOST));
+				CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_SUBSCR_ATTACH_LOST);
 			}
 			continue;
 		}
@@ -1054,10 +1065,10 @@ struct hnbgw_cnlink *hnbgw_cnlink_select(struct hnbgw_context_map *map)
 					cnpool->peer_name, cnlink->nr);
 			} else {
 				LOG_NRI(LOGL_INFO, "NRI match selects %s %d\n", cnpool->peer_name, cnlink->nr);
-				rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_SUBSCR_KNOWN));
+				CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_SUBSCR_KNOWN);
 				if (map->l3.is_emerg) {
-					rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_EMERG_FORWARDED));
-					rate_ctr_inc(rate_ctr_group_get_ctr(cnpool->ctrs, CNPOOL_CTR_EMERG_FORWARDED));
+					CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_EMERG_FORWARDED);
+					CNPOOL_CTR_INC(cnpool, CNPOOL_CTR_EMERG_FORWARDED);
 				}
 				return cnlink;
 			}
@@ -1094,22 +1105,22 @@ struct hnbgw_cnlink *hnbgw_cnlink_select(struct hnbgw_context_map *map)
 	 * them are usable -- wrap to the start. */
 	cnlink = round_robin_next ? : round_robin_first;
 	if (!cnlink) {
-		rate_ctr_inc(rate_ctr_group_get_ctr(cnpool->ctrs, CNPOOL_CTR_SUBSCR_NO_CNLINK));
+		CNPOOL_CTR_INC(cnpool, CNPOOL_CTR_SUBSCR_NO_CNLINK);
 		if (map->l3.is_emerg)
-			rate_ctr_inc(rate_ctr_group_get_ctr(cnpool->ctrs, CNPOOL_CTR_EMERG_LOST));
+			CNPOOL_CTR_INC(cnpool, CNPOOL_CTR_EMERG_LOST);
 		return NULL;
 	}
 
 	LOG_MAP(map, DCN, LOGL_INFO, "CN link round-robin selects %s %d\n", cnpool->peer_name, cnlink->nr);
 
 	if (is_null_nri)
-		rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_SUBSCR_REATTACH));
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_SUBSCR_REATTACH);
 	else
-		rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_SUBSCR_NEW));
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_SUBSCR_NEW);
 
 	if (map->l3.is_emerg) {
-		rate_ctr_inc(rate_ctr_group_get_ctr(cnlink->ctrs, CNLINK_CTR_CNPOOL_EMERG_FORWARDED));
-		rate_ctr_inc(rate_ctr_group_get_ctr(cnpool->ctrs, CNPOOL_CTR_EMERG_FORWARDED));
+		CNLINK_CTR_INC(cnlink, CNLINK_CTR_CNPOOL_EMERG_FORWARDED);
+		CNPOOL_CTR_INC(cnpool, CNPOOL_CTR_EMERG_FORWARDED);
 	}
 
 	/* A CN link was picked by round-robin, so update the next round-robin nr to pick */
@@ -1130,6 +1141,43 @@ char *cnlink_sccp_addr_to_str(struct hnbgw_cnlink *cnlink, const struct osmo_scc
 }
 
 static const struct rate_ctr_desc cnlink_ctr_description[] = {
+	[CNLINK_CTR_RANAP_RX_UDT_RESET] = {
+		"ranap:rx:udt:reset",
+		"RANAP Unitdata RESET messages received"
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_RESET_ACK] = {
+		"ranap:rx:udt:reset_ack",
+		"RANAP Unitdata RESET ACK messages received",
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_PAGING] = {
+		"ranap:rx:udt:paging",
+		"RANAP Unitdata PAGING messages received",
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_UNKNOWN] = {
+		"ranap:rx:udt:unknown",
+		"Unknown RANAP Unitdata messages received",
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_UNSUPPORTED] = {
+		"ranap:rx:udt:unsupported",
+		"Unsupported RANAP Unitdata messages received",
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_OVERLOAD_IND] = {
+		"ranap:rx:udt:overload_ind",
+		"RANAP Unitdata Overload Indications received",
+	},
+	[CNLINK_CTR_RANAP_RX_UDT_ERROR_IND] = {
+		"ranap:rx:udt:error_ind",
+		"RANAP Unitdata Error Indications received",
+	},
+
+	[CNLINK_CTR_RANAP_TX_UDT_RESET] = {
+		"ranap:tx:udt:reset",
+		"RANAP Unitdata RESET messages transmitted",
+	},
+	[CNLINK_CTR_RANAP_TX_UDT_RESET_ACK] = {
+		"ranap:tx:udt:reset_ack",
+		"RANAP Unitdata RESET ACK messages transmitted",
+	},
 
 	/* Indicators for CN pool usage */
 	[CNLINK_CTR_CNPOOL_SUBSCR_NEW] = {
